/*
 * application_layer.java
 * Douglas S. J. De Couto
 * 1 September 1999
 */

/*
 * Copyright (C) 1999 Douglas S. J. De Couto
 * <decouto@lcs.mit.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package decouto.garble;

import java.util.Vector;

/** Communicates with GPS via the Garmin link layer to transfer data
    to and from the GPS and control the GPS.
    @author Douglas S. J. De Couto <a href="mailto:decouto@lcs.mit.edu">decouto@lcs.mit.edu</a>
    @version 1 September 1999 0.1 
    @see decouto.garble.link_layer */
class application_layer {

  /** Default constructor -- creates an unuseable application layer
      object */
  public application_layer() {
    m_init = false;
    m_ll = null;
    m_got_capability_protocol = false;
  }

  /** Link layer constructor -- you should use this instead of the
      default.
      @param ll An initialized garmin link layer object */
  public application_layer(link_layer ll) {
    if (ll != null) {
      m_ll = ll;
      m_init = true;
    }
    else {
      m_ll = null;
      m_init = false;
    }
    m_got_capability_protocol = false;
  }

  /** Turn off the GPS.
      @exception not_possible Thrown if there is an error exchanging
      packets with the GPS. 
      @exception timeout Thrown if the GPS does not respond in time. */
  public void turn_off_gps() throws not_possible, timeout {
    check_init();

    packet p = new packet();
    p.id = packet.pid_command_data;
    p.sz = 2;
    util.short_to_bytes(command.cmnd_turn_off_pwr, p.data);
    m_ll.put_packet(p);
  }

  /** Get GPS product info.
      @return product_data structure.
      @exception not_possible Thrown if there is an error exchanging
      packets with the GPS. 
      @exception timeout Thrown if the GPS does not respond in time. 
      @see decouto.garble.product_data */
  public product_data
  get_product_data() throws not_possible, timeout {
    check_init();

    packet p = new packet(packet.pid_product_rqst, 0);
    m_ll.put_packet(p);
    m_ll.get_packet(p);

    if (p.id != packet.pid_product_data) {
      throw new not_possible("GPS did not reply with product data packet");
    }
    if (p.sz < 5) {
      throw new not_possible("Product data packet is too small");
    }
	
    product_data pd = new product_data();
    pd.product_id = util.bytes_to_short(p.data);
    pd.software_version = util.bytes_to_short(p.data, 2);
    /* find length of prod description string */
    int desc_sz = 0;
    for (int i = 4; i < p.sz && p.data[i] != 0; i++) {
      desc_sz++;
    }
    pd.product_description = new String(p.data, 4, desc_sz);

    /* should be smart to see if we need to fetch capability protocol
    info, but for now we will just depend upon timeout.  bad though
    because maybe not all physical layers will support timeout ops.
    Will change especially when i switch it to using streams... */
    try {
      m_ll.get_packet(p);
    }
    catch (timeout ex) {
      // timeout indicates GPS is not sending capability protocol info
      // (A001)
      m_got_capability_protocol = false;
      return pd;
    }
    catch (not_possible ex) {
      m_got_capability_protocol = false;
      return pd;
    }

    // receive capability protocol
    m_got_capability_protocol = true;
    if (p.id != packet.pid_protocol_array) {
      throw new not_possible("Product data packet followed by some packet besides capability protocol");
    }
    if ((p.sz % 3) != 0) {
      throw new not_possible("Capability protocol packet is strange size");
    }
    
    Vector v = new Vector (p.sz / 3);
    for (int i = 0; i < p.sz; i +=3) {
      protocol_data prot_data = new protocol_data();
      prot_data.tag = p.data[i];
      prot_data.data = (util.bytes_to_short(p.data, i + 1) & 0xFFff);
      v.addElement(prot_data);
    }
    m_prot_data_arr = v;
    return pd;
  }

  /** Returns the waypoints stored in the GPS.
      @return Vector of basic_waypt objects.
      @exception not_possible Thrown if there is an error exchanging
      packets with the GPS. 
      @exception timeout Thrown if the GPS does not respond in time. 
      @see decouto.garble.basic_waypt */
  public Vector /* [basic_waypt] */
  get_waypoints() throws not_possible, timeout {
    check_init();

    packet p = new packet(packet.pid_command_data, 2);
    util.short_to_bytes(command.cmnd_transfer_wpt, p.data);
    m_ll.put_packet(p);

    m_ll.get_packet(p);
    if (p.id != packet.pid_records) {
      throw new not_possible("GPS did not reply with start of records data packet for waypoint data");
    }
    if (p.sz != 2) {
      throw new not_possible("Waypoint data start of records packet is wrong size");
    }

    short num_recs = util.bytes_to_short(p.data);
    Vector v = new Vector(num_recs);

    for (short i = 0; i< num_recs; i++) {
      basic_waypt waypt = new basic_waypt();
      v.addElement(waypt);

      m_ll.get_packet(p);
      if (p.id != packet.pid_wpt_data) {
	throw new not_possible("GPS did not send waypoint data packet");
      }
      if (p.sz < (6 + 8 + 4 + 40)) {
	throw new not_possible("Waypoint data packet is too small");
      }

      semicircle_pos pos = new semicircle_pos();
      waypt.pos = pos;
      pos.lat = util.bytes_to_int(p.data, 6);
      pos.lon = util.bytes_to_int(p.data, 10);

      waypt.id = new String(p.data, 0, 6);
      waypt.comment = new String(p.data, 18, 40);
    }

    m_ll.get_packet(p);
    if (p.id != packet.pid_xfer_cmplt) {
      throw new not_possible("GPS did not terminate waypoint data with end of records packet");
    }
    if (util.bytes_to_short(p.data) != command.cmnd_transfer_wpt) {
      throw new not_possible("End of records packet does not match original waypoint transfer command");
    }
    
    return v;
  }
  /** Check that the application layer object is initialized.
      @exception not_possible Thrown if object is uninitialized. */
  private void check_init() throws not_possible {
    if (!m_init) {
      throw new not_possible("application_layer not initialized");
    }
  }

  /** True iff object is initialized */
  private boolean m_init;
  /** Reference to underlying link layer; must be non-null.
      @see decouto.garble.link_layer */
  private link_layer m_ll;
  /** true iff we have read the capability protocol data from GPS */
  private boolean m_got_capability_protocol;
  /** If non-null, a Vector of protocol_data objects describing the
      GPS's interface capabilities.  The vector may contain duplicates
      and be unsorted.
      @see java.util.Vector
      @see decouto.garble.protocol_data */
  Vector m_prot_data_arr = null; 
}





