/*
 * link_layer.java
 * Douglas S. J. De Couto
 * 26 August 1999
 */

/*
 * Copyright (C) 1999 Douglas S. J. De Couto
 * <decouto@lcs.mit.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package decouto.garble;

/** Send data packets to and from a device implementing the Garmin
    L000/1 protocol.  Based on information in ``Garmin GPS Interface
    Specification'' , from Garmin tech support, May 6 1998.  Handles
    handshaking, checksumming, and data escaping.  Handle
    checksumming, acknowledgements, and retransmits.  Used by the
    Garmin application layer: decouto.garble.application_layer.
    Relies upon decouto.garble.physical_layer for actual byte data
    transmission.
    @author Douglas S. J. De Couto <a href="mailto:decouto@lcs.mit.edu">decouto@lcs.mit.edu</a>
    @version 26 August 1999 0.1 
    @see decouto.garble.application_layer
    @see decouto.garble.physical_layer 
    @see decouto.garble.packet */
class link_layer {

  /** Default constructor -- creates an unuseable link layer
      object */
  public link_layer() {
    m_init = false;
  }

  /** Physical layer constructor -- you should use this instead of the
      default.
      @param pl An initialized garmin physical layer object.
      @param retries The number of times to retransmit an
      unacknowledged packet before giving up. */
  public link_layer(physical_layer pl, int retries) {
    m_pl = pl;
    m_retries = retries;
    m_pkt = new packet();
    m_init = true;
  }

  /** Send a NAK packet to the device.
      @param id The id of the packet which we are NAK-ing.
      @exception not_possible Thrown if there is an error sending or
      receiving data with the GPS.
      @exception timeout Thrown if the GPS does not respond in time. */
  private void send_nak(byte id) throws not_possible, timeout {
    if (debug.debug_enabled) {
      debug.println("link_layer::send_nak -- sending NAK");
    }

    m_pl.putbyte(dle);
    m_pl.putbyte(packet.pid_nack_byte);
    m_pl.putbyte((byte) 2);
    // indicate bad packet id with 16 bits; but all received packets in this version of protocol only have 8 bit ids
    m_pl.putbyte(id); 
    m_pl.putbyte((byte) 0);
    byte nack_cksum = (byte) (256 - (((packet.pid_nack_byte & 0xff) + 2 + (id & 0xff)) & 0xff));
    if (nack_cksum == dle) {
      m_pl.putbyte(dle);
    }
    m_pl.putbyte(nack_cksum);
    m_pl.putbyte(dle);
    m_pl.putbyte(etx);
  }
  
  /** Receive a packet from the GPS.  
      @return The ID of the received packet.
      @param p A packet object, whose fields are filled in by this method.
      @exception not_possible Thrown if there is an error sending or
      receiving data with the GPS.
      @exception timeout Thrown if the GPS does not respond in time. */
  byte get_packet(packet p) throws not_possible, timeout {
    if (!m_init) {
      throw new not_possible("The Garmin link layer is not initialized");
    }
    try {
      for (int curr_try = 1; curr_try <= m_retries; curr_try++) {
	int bytes_scanned = 0;
	while (true) {
	  if (bytes_scanned > max_random_bytes) {
	    throw new not_possible("Too many random bytes encountered while waiting for a packet");
	  }
	  /* scan until we get first byte of a packet... */
	  p.id = m_pl.getbyte();
	  bytes_scanned++;
	  if (p.id == dle) {
	    // could be it, but could be dle in chunk of leftover packet...
	    p.id = m_pl.getbyte();
	    bytes_scanned++;
	    // valid packet start dle is never followed by dle or etx
	    if (p.id == dle) {
	      // ``dle stuffing'' in packet data, ignore...
	      continue;
	    }
	    if (p.id == etx) {
	      // end of old packet frame, ignore...
	      continue;
	    } 
	    break;
	  }
	}
	
	byte b = m_pl.getbyte();
	if (b == dle) {
	  b = m_pl.getbyte();
	  if (b != dle) {
	    if (debug.debug_enabled) {
	      debug.println("link_layer::get_packet -- missing dle stuffing in packet size");
	    }
	    send_nak(p.id);
	  }
	}
	p.sz = (b & 0xff);
	
	for (int i = 0; i < p.sz; i++) {
	  p.data[i] = m_pl.getbyte();
	  if (p.data[i] == dle) {
	    // check for and ignore correct dle stuffing byte
	    b = m_pl.getbyte();
	    if (b != dle) {
	      if (debug.debug_enabled) {
		debug.println("link_layer::get_packet -- missing dle stuffing in packet data");
	      }
	      send_nak(p.id);
	    }
	  }
	}
	
	byte packet_cksum = m_pl.getbyte();
	if (packet_cksum == dle) {
	  packet_cksum = m_pl.getbyte();
	  if (packet_cksum != dle) {
	    if (debug.debug_enabled) {
	      debug.println("link_layer::get_packet -- missing dle stuffing in packet cksum");
	    }
	    send_nak(p.id);
	  }
	}
	
	if (p.calc_cksum() != packet_cksum) {
	  if (debug.debug_enabled) {
	    debug.println("link_layer::get_packet -- bad checksum");
	  }
	  send_nak(p.id);
	}
	
	byte b1, b2;
	b1 = m_pl.getbyte();
	b2 = m_pl.getbyte();
	if (b1 != dle || b2 != etx) {
	  if (debug.debug_enabled) {
	    debug.println("link_layer::get_packet -- bad packet framing");
	    debug.println("link_layer::get_packet -- id is " + Integer.toHexString(p.id & 0xff));
	    debug.println("link_layer::get_packet -- sz is " + p.sz);
	    debug.println("link_layer::get_packet -- data is: ");
	    for (int i = 0; i < p.sz; i++) {
	      debug.print(Integer.toHexString(p.data[i] & 0xff) + " ");
	    }
	    debug.println("");
	    debug.println("link_layer::get_packet -- cksum is " + Integer.toHexString(packet_cksum & 0xff));
	    debug.println("link_layer::get_packet -- dle byte is " + Integer.toHexString(b1 & 0xff));
	    debug.println("link_layer::get_packet -- etx byte is " + Integer.toHexString(b2 & 0xff));
	  }
	  send_nak(p.id);
	}
	
	// if we got this far, we got the packet ok, so send ACK
	m_pl.putbyte(dle);
	m_pl.putbyte(packet.pid_ack_byte);
	m_pl.putbyte((byte) 2);
	m_pl.putbyte(p.id);
	m_pl.putbyte((byte) 0); // msb of 16 bit id must be 0, since all ids are only one byte... 
	byte ack_cksum = (byte) (256 - (((packet.pid_ack_byte & 0xff) + 2 + (p.id & 0xff)) & 0xff));
	if (ack_cksum == dle) {
	  m_pl.putbyte(dle);
	}
	m_pl.putbyte(ack_cksum);
	m_pl.putbyte(dle);
	m_pl.putbyte(etx);
	
	return  p.id; 
      } // for curr_try...
      throw new not_possible("Too many bad packets encountered  while waiting for a good packet");
    } // try
    catch (not_possible ex) {
      throw ex;
    }
    catch (timeout ex) {
      throw ex;
    }
  }

  /** Send a packet to the GPS.  
      @param p A packet object, whose fields must be filled in
      appropriately.
      @exception not_possible Thrown if there is an error sending or
      receiving data with the GPS.
      @exception timeout Thrown if the GPS does not respond in time. */
  void put_packet(packet p) throws not_possible, timeout {
    // sends out the data in packet p, using supplied 
    // size and id.
    if (!m_init) {
      throw new not_possible("The garmin link layer is not initialized");
    }
    
    try {
      int curr_try = 0;
      while (true) {
	curr_try++;
	if (curr_try > m_retries) {
	  throw new not_possible("Too many retries (" + curr_try + ") while sending packet");
	}
	
	// try to send packet...			
	// put packet header
	m_pl.putbyte(dle);
	m_pl.putbyte(p.id);
	if ((byte) p.sz == dle) {
	  m_pl.putbyte(dle);
	}
	m_pl.putbyte((byte) p.sz);
	
	// put data
	for (int i = 0; i < p.sz; i++) {
	  if (p.data[i] == dle) {
	    m_pl.putbyte(dle);
	  }
	  m_pl.putbyte(p.data[i]);
	}
	
	// trailer
	byte cksum = p.calc_cksum();
	m_pl.putbyte(cksum);
	if (cksum == dle) {
	  m_pl.putbyte(dle);
	}
	m_pl.putbyte(dle);
	m_pl.putbyte(etx);
	
	// wait for ack.
	// assume all packets are delivered in order, and that an ack/nak follows each 
	// packet received by the gps.  i.e. next incoming packet is our ACK or NACK.
	byte rcv_id;
	int bytes_scanned = 0;
	while (true) {
	  if (bytes_scanned > max_random_bytes) {
	    throw new not_possible("Too many random bytes encountered while waiting for ack packet");
	  }
	  rcv_id = m_pl.getbyte();
	  bytes_scanned++;
	  if (rcv_id == dle) {
	    rcv_id = m_pl.getbyte();
	    bytes_scanned++;
	    if (rcv_id == dle) {
	      continue;
	    }
	    if (rcv_id == etx) {
	      continue;
	    } 
	    break;
	  }
	  /* should check for a timeout here, in case of garbage data stream... */
	}
	m_pkt.id = rcv_id;

	debug.println("starting to read ack...");

	byte rcv_sz = m_pl.getbyte();
	if (rcv_sz == dle) {
	  rcv_sz = m_pl.getbyte();
	  if (rcv_sz != dle) {
	    if (debug.debug_enabled) {
	      debug.println("link_layer::put_packet -- missing dle stuffing in reply packet size");
	    }
	    send_nak(rcv_id);
		continue;
	  }
	}
	m_pkt.sz = (rcv_sz & 0xff);
	
	for (int i = 0; i < m_pkt.sz; i++) {
	  byte b = m_pl.getbyte();
	  if (b == dle) {
	    b = m_pl.getbyte();
	    if (b != dle) {
	      if (debug.debug_enabled) {
		debug.println("link_layer::put_packet -- missing dle stuffing in reply packet data");
	      }
	      send_nak(rcv_id);	
			continue;		
	    }
	  }
	  m_pkt.data[i] = b;
	}
	
	byte rcv_cksum = m_pkt.calc_cksum();
	byte b = m_pl.getbyte();
	if (b == dle) {
	  b = m_pl.getbyte();
	  if (b != dle) {
	    if (debug.debug_enabled) {
	      debug.println("link_layer::put_packet -- missing dle stuffing in reply packet cksum");
	    }
	    send_nak(rcv_id);
		continue;
	  }
	}
	if (b != rcv_cksum) {
	  if (debug.debug_enabled) {
	    debug.println("link_layer::put_packet -- bad cksum in reply packet");
		debug.println("link_layer::put_packet -- got " + Integer.toHexString(b & 0xff) + 
						" for packet: " + m_pkt.hex_dump());
	  }
	  send_nak(rcv_id);
		continue;
	}
	byte b1, b2;
	b1 = m_pl.getbyte();
	b2 = m_pl.getbyte();			
	if (b1 != dle || b2 != etx) {
	  if (debug.debug_enabled) {
	    debug.println("link_layer::put_packet -- bad packet framing in reply packet");
	    debug.println("link_layer::put_packet -- id is " + rcv_id);
	    debug.println("link_layer::put_packet -- sz is " + rcv_sz);
	    debug.println("link_layer::put_packet -- data is: ");
	    for (int i = 0; i < rcv_sz; i++) {
	      debug.print(Integer.toHexString(m_pkt.data[i] & 0xff) + " ");
	    }
	    debug.println("");
	    debug.println("link_layer::put_packet -- cksum is " + Integer.toHexString(rcv_cksum & 0xff));
	    debug.println("link_layer::put_packet -- dle byte is " + Integer.toHexString(b1 & 0xff));
	    debug.println("link_layer::put_packet -- etx byte is " + Integer.toHexString(b2 & 0xff));
	  }
	  send_nak(rcv_id);
		continue;
	}
	if (rcv_id == packet.pid_nack_byte || rcv_id == packet.pid_ack_byte) {
	  // check packet size and data
	  if (m_pkt.sz != 2) {
	    if (debug.debug_enabled) {
	      debug.println("link_layer::put_packet -- ack/nack packet has bad size");
	    }
	    send_nak(rcv_id);
		continue;
	  }
	  if (m_pkt.data[0] != p.id) {
	    // this ack/nack is not for what we just transmitted, resend our packet.
	    if (debug.debug_enabled) {
	      debug.println("link_layer::put_packet -- ack/nack packet is for some other packet id");
	    }
	    continue;
	  }
	  if (rcv_id == packet.pid_ack_byte) {
	    break;
	  }
	  if (rcv_id == packet.pid_nack_byte) {
	    if (debug.debug_enabled) {
	      debug.println("link_layer::put_packet -- got nack, retransmitting");
	    }
	    continue; // retransmit...
	  }
	}
	else {
	  // not an ack or nack reply, send our packet again...
	  if (debug.debug_enabled) {
	    debug.println("link_layer::put_packet -- got some packet besides ack/nack (" 
			  + rcv_id + "), retransmitting");
	    if (rcv_id == 10) { // what was this case for? 10 is: pid_command_data
	      short cmndval = (short) ((m_pkt.data[1] << 8) & 0xff00); // high byte
	      cmndval |= (short) (m_pkt.data[0] & 0xff); // low byte
	      debug.println("cmnd == " + Integer.toHexString(cmndval & 0xFFff));
	    } 
	  }
	  continue;
	}
	// loop again to retransmit original packet
      } // while
    } // try
    catch (not_possible ex) {
      throw ex;
    }
    catch (timeout ex) {
      throw ex;
    }
  }

  /** Byte value starting and ending packet frames. */
  private static final byte dle = 0x10;
  /** Byte value ending packet frames. */
  private static final byte etx = 0x03;
  /** Maximum number of random bytes to scan looking for the start of
      a packet frame.  Should be large enough to skip over the largest
      possible packet, in case we just missed the beginning of a
      packet. */
  private static final int max_random_bytes = 256; 

  /** true iff object is initialized */
  private boolean m_init = false;
  /** Reference to underlying physical layer; must be non-null. 
      @see decouto.garble.physical_layer */
  private physical_layer m_pl = null;
  
  /** scratch packet used for receiving ACKs/NAKs, etc. 
      @see decouto.garble.packet */
  private packet m_pkt; 
  /** number of times to retry sending a packet when we do not receive
      an ACK. */
  private int m_retries;
}
